///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Program Name:  CCrypt.c                                                   //
//      Authors:  Jim Aspnes (Copyright 1997 by Jim Aspnes)                  //
//                Mike McCollister (UI Enhancements)                         //
//     Language:  PRC-Tools 2.3                                              //
//                Palm OS 5 SDK R3                                           //
//                PilRC 3.2                                                  //
//                HandEra Header Files from SDK 1.04                         //
//                Sony Header Files from SDK 5.0 version 1.2                 //
//  Description:  Clipboard encryptor.                                       //
//        Notes:  Distributed under the terms of the GNU General Public      //
//                License version 2.0.                                       //
//                                                                           //
//                Some parts of this program are derived from:               //
//                   hello-world.c:  Simple test program                     //
//                   (c) 1996, 1997 D. Jeff Dionne                           //
//                   This is free software, under the GNU public license v2. //
//===========================================================================//
//  Version: 2.1.1                                                           //
//     Date: Unknown Date (1997)                                             //
//   Author: Jim Aspnes                                                      //
// Comments: Version available on the Internet.                              //
//---------------------------------------------------------------------------//
//  Version: 2.2.0                                                           //
//     Date: 07-27-1999                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Mostly GUI changes and cleanup (menu reordered, buttons added,  //
//           etc.).                                                          //
//                                                                           //
//           Clear now clears the clipboard as well if the user wants.       //
//                                                                           //
//           The fields are cleared when the power is turned off.            //
//---------------------------------------------------------------------------//
//  Version: 2.2.1                                                           //
//     Date: 02-26-2000                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Removed CLICK from prevFieldChr since Palm OS 3.3 now clicks    //
//           when the previous field character is entered.                   //
//                                                                           //
//           Default button specified in dialogs when appropriate.           //
//                                                                           //
//           To avoid any errors using the debug version of the emulator,    //
//           the code was modified so that there is now a MainEventHandler.  //
//           This along with some other modifications removed many errors.   //
//                                                                           //
//           Added FrmCloseAllForms to prevent memory leak.                  //
//                                                                           //
//           Cleaned up the Makefile some more.  Also put the source files   //
//           in the src directory.                                           //
//                                                                           //
//           Clear at power off now works with PalmOS 3.2 and greater.  With //
//           older versions of the OS, holding down the power button to turn //
//           on the backlight will clear all fields.                         //
//                                                                           //
//           Better about box form.                                          //
//                                                                           //
//           Switched to using PRC-Tools 2.0 with Palm OS 3.5 header files.  //
//           Because of this, I had to make the following changes:           //
//              * Use new types (i.e. UInt8 instead of Byte) when            //
//                applicable.                                                //
//              * New header file format.                                    //
//              * The pragma pack(2) is no longer needed                     //
//              * Modified the Makefile so that it will work with PRC-Tools  //
//                2.0.                                                       //
//                                                                           //
//           Now have color icons.                                           //
//---------------------------------------------------------------------------//
//  Version: 2.3.0                                                           //
//     Date: 04-01-2000                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Added a mask key option.                                        //
//                                                                           //
//           Cleaned up the Makefile some.                                   //
//---------------------------------------------------------------------------//
//  Version: 2.3.1                                                           //
//     Date: 04-15-2000                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses PilRC 2.5b4.                                           //
//                                                                           //
//           Icon now has proper transparency.                               //
//                                                                           //
//           "Get Info" icon shows up in the command bar on Palm OS 3.5 or   //
//           greater.                                                        //
//                                                                           //
//           Added scrollbar for Palm OS 2 or greater devices.               //
//                                                                           //
//           Added font selection for data field.                            //
//                                                                           //
//           Maximum characters for the data field is up from 1000 to 4096.  //
//                                                                           //
//           Removed MenuEraseStatus(0) since it is no longer needed with    //
//           the buttons at the bottom of the screen.                        //
//                                                                           //
//           After a decrypt, the cursor is placed to the top of the data    //
//           field.                                                          //
//---------------------------------------------------------------------------//
//  Version: 2.3.2                                                           //
//     Date: 06-11-2000                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses PilRC 2.5b6.                                           //
//                                                                           //
//           Minor GUI changes.                                              //
//---------------------------------------------------------------------------//
//  Version: 2.3.3                                                           //
//     Date: 11-24-2000                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses PilRC 2.6.                                             //
//                                                                           //
//           Updated contact information.                                    //
//---------------------------------------------------------------------------//
//  Version: 2.3.4                                                           //
//     Date: 04-12-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses PilRC 2.7b.                                            //
//                                                                           //
//           Made some minor UI changes.  The menu now shows "About CCrypt"  //
//           instead of just "About".                                        //
//                                                                           //
//           Made some more type changes to be better suited for Palm OS     //
//           3.5 header files.                                               //
//                                                                           //
//           The Makefile now defaults to using WinZip for compressing the   //
//           source and distribution files.                                  //
//---------------------------------------------------------------------------//
//  Version: 2.4.0                                                           //
//     Date: 05-16-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses PilRC 2.8.                                             //
//                                                                           //
//           Now uses Palm OS 4 header files.                                //
//                                                                           //
//           Now uses PRC-Tools 2.0.92.                                      //
//                                                                           //
//           Added an error prevention section.                              //
//                                                                           //
//           Makefile modifed so bring out more potential errors.  Made      //
//           adjustments to the code because of this.                        //
//                                                                           //
//           Changed to version 2.4.0.                                       //
//                                                                           //
//           Fixed a small memory leak bug (invalid field length) on exit.   //
//                                                                           //
//           Moved the scroll bar over the right right one pixel and made    //
//           the data field one pixel wider.  This makes CCrypt the same     //
//           dimensions as the built-in memo pad.                            //
//---------------------------------------------------------------------------//
//  Version: 2.5.0                                                           //
//     Date: 05-26-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now supports the HandEra 330.                                   //
//           * Includes support for minimizing/maximizing of the Virtual     //
//             Silkscreen.                                                   //
//           * Supports the HandEra 330 system font set.                     //
//           * Supports other 160x160 Palm-Powered devices (Palm IIIxe,      //
//             m500, etc).                                                   //
//           * Includes large and medium icons (icon IDs 2000, 2001, 2002).  //
//           * Appears as High-Res in Application Preference Panel.          //
//---------------------------------------------------------------------------//
//  Version: 2.5.1                                                           //
//     Date: 06-13-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Fixed a bug that prevented the page up and page down keys from  //
//           scrolling the data the proper number of lines.                  //
//                                                                           //
//           Added color icons for the HandEra.  While no color HandEra      //
//           device exists yet, I decided to put these in now.               //
//---------------------------------------------------------------------------//
//  Version: 2.5.2                                                           //
//     Date: 06-23-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Added the ASCII Armor checkbox so that the ASCII Armor method   //
//           can be used by default.                                         //
//---------------------------------------------------------------------------//
//  Version: 2.5.9                                                           //
//     Date: 09-15-2001                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Moved the medium icon up one pixel.                             //
//---------------------------------------------------------------------------//
//  Version: 2.6.0                                                           //
//     Date: 07-09-2002                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now uses Palm OS 5.0 SDK.  Because of this                      //
//           DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS is no longer        //
//           required.                                                       //
//                                                                           //
//           Due to using PilRc 2.9p5, the default button is now correctly   //
//           specified on the alert forms, thus preventing a possible memory //
//           leak.                                                           //
//                                                                           //
//           Modified the Makefile so that it uses the -ro option to         //
//           simplify the compile.                                           //
//                                                                           //
//           Icons are now compressed.                                       //
//                                                                           //
//           Now requires Palm OS 2.0 or greater.                            //
//                                                                           //
//           Removed V10 calls.                                              //
//                                                                           //
//           On the HandEra 330, when the graffiti area is maximized the     //
//           cursor will be moved if need be.                                //
//                                                                           //
//           Removed "ASCII Armor Clipboard" from the menu as it was         //
//           redundant and confusing.  The ASCII Armor option is still       //
//           available as a checkbox at the top of the screen.               //
//                                                                           //
//           No longer forces the screen orientation to be "normal" when     //
//           started on a HandEra 330.                                       //
//                                                                           //
//           Added large icons for Palm OS 5.0 devices.                      //
//                                                                           //
//           Verified compatible with Palm OS 5.0 devices.                   //
//                                                                           //
//           Added more help text to the "Get Info" menu item.               //
//---------------------------------------------------------------------------//
//  Version: 2.61                                                            //
//     Date: 02-25-2003                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Now compiles with PRC-Tools 2.2 and PilRC 2.9p9.                //
//                                                                           //
//           Changed the "Get Info" menu item to "Info...".                  //
//                                                                           //
//           MenuCmdBarAddButton now uses menuCmdBarOnRight instead of       //
//           menuCmdBarLeft.                                                 //
//                                                                           //
//           Changed the "L" menu shortcut to "A" for Graffiti 2             //
//           compatibility.                                                  //
//                                                                           //
//           Added a menu items to beam and send "CCrypt".                   //
//---------------------------------------------------------------------------//
//  Version: 2.70                                                            //
//     Date: 12-18-2004                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Changed support e-mail address to support@MikeMcCollister.com.  //
//                                                                           //
//           Changed C coding spacing style.  Also clean up the code.        //
//                                                                           //
//           Now uses PRC-Tools 2.3.                                         //
//                                                                           //
//           Now uses Palm OS 5 SDK R3.                                      //
//                                                                           //
//           Now uses PilRC 3.2.                                             //
//                                                                           //
//           Added 108 dpi icons.                                            //
//                                                                           //
//           Removed the color HandEra icons since there will never be such  //
//           a device, and all they were doing was wasting space.            //
//                                                                           //
//           Renamed the bitmap files to better describe their densities.    //
//                                                                           //
//           Added default button IDs that were missing in a couple of       //
//           forms.                                                          //
//                                                                           //
//           In the edit menu, "Graffiti Help" will now display              //
//           "Graffiti 2 Help" for devices that have Graffiti 2.             //
//                                                                           //
//           Much nicer Makefile and better organized outputs.               //
//                                                                           //
//           Now takes advantage of the resizable graffiti area all devices  //
//           that have such a feature.                                       //
//---------------------------------------------------------------------------//
//  Version: 2.71                                                            //
//     Date: 10-20-2006                                                      //
//   Author: Michael J. McCollister                                          //
// Comments: Makefile now uses zip (Info-ZIP) in stead of WinZip.            //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

///////////////////
// Include Files //
///////////////////

#include "CCrypt.h"
#include "CCrypt.rch"

/////////////
// Defines //
/////////////

#define APP_ID			'arQx'
#define APP_VER			271
#define PREF_VER		260
#define PREF_ID                 0

#define BEEP			SndPlaySystemSound(sndInfo)
#define DEBUG_BEEP		BEEP
#define CLICK			SndPlaySystemSound(sndClick)

#define BREAKPOINT() 		{ volatile asm("trap #1"); }

// we'll put the tick count at the beginning of all plaintexts
// so that the same plaintext doesn't yield the same ciphertext every time
#define HEADER			(sizeof(UInt32))
#define MAX_PLAINTEXT_LEN	(MAXLEN - HEADER)

#ifndef SetBits
#define SetBits(b, len)         (((1U << ((len) - 1)) - 1U +  \
                                  (1U << ((len) - 1))) << (b))
#endif

// pinMaxConstraintSize is set to the max value of a Coord such that it can
// grow as big as the hardware supports
#ifndef pinMaxConstraintSize
#define pinMaxConstraintSize 	SetBits(0, (sizeof(Coord) * 8) - 1)
#endif

////////////////
// Structures //
////////////////

typedef struct
{
   Boolean asciiArmorKey;
   Boolean maskKey;
   FontID font;
} Data;

//////////////////////
// Global Variables //
//////////////////////

Data data;
Boolean realPowerOffAvail;
Boolean bVgaExists = false;
RectangleType oldWinRect = {{0, 0}, {160, 160}};
VgaRotateModeType currentRotateMode;
Boolean bCollapsibleGraffiti = false;

/////////////////////////
// Function Prototypes //
/////////////////////////

// Common Palm Functions
static UInt16 CheckOsVersion(Boolean bDisplayAlert);
static UInt16 StartApplication(void);
static void LoadData(void);
static void EventLoop(void);
static Boolean AppHandleEvent(EventPtr event);
static void StopApplication(void);
static struct cbuf *CbufNew(void);
static void CbufFree(struct cbuf *c);
static void ClearField(FieldPtr field);
static struct cbuf *GetClipboard(void);
static UInt16 GetCurrentFieldId(void);
static FieldPtr GetFieldOfActive(UInt16 field);
static FieldPtr GetFocusObjectPtr(void);
static UInt16 GetROMVer(UInt16 *ROMVerMajor, UInt16 *ROMVerMinor);
static void *Id2Ptr(UInt16 wID);
static Err SendMe(Char *prefix);
static Err SendDatabase(UInt16 cardNo, LocalID dbID, Char *nameP,
                        Char *descriptionP);
static Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP);

// Main Form Functions
static Boolean MainEventHandler(EventPtr e);
static void HandleNextPrevField(UInt16 keyDown);
static void SetKey(key k);
static void DoEncryptCmd(void);
static Int16 SanityCheck(struct cbuf *v);
static void DoDecryptCmd(void);
static void DoArmorCmd(Boolean calledFromEncrypt);
static void DoClearCmd(Boolean askAboutClipboard);
static Boolean ButtonCommand(UInt16 command);
static Boolean MenuCommand(UInt16 command);
static void Scroll(Boolean scrolldown);
static void ChangeToGraffiti2MenuHelp(void);

// About Form Functions
static void DisplayAboutForm(UInt16 rsdID);

// Scrollbar Functions
static void EditViewUpdateScrollBar(void);
static void EditViewScroll(Int16 linesToScroll, Boolean updateScrollbar);

// HandEra & DIA Resize Functions
static void SizeForm(FormPtr pForm);
static UInt16 GetGSIIndex(FormPtr pForm);
static void ResizeObject(FormPtr pForm, UInt16 objectIndex,
                         Coord xMove, Coord yMove,
                         Coord xExp, Coord yExp, Boolean bDraw,
                         Boolean bMultiLineField);
static void CollapseSetUserState(FormPtr pForm, Boolean bForce);
static Boolean GetCollapseGraffiti(Boolean *pbPS, Boolean *pbSony);
static Err MySysNotifyDisplayChangeEvent(SysNotifyParamType *np);

///////////////////
// Main Function //
///////////////////

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP __attribute__ ((unused)),
                 UInt16 launchFlags __attribute__ ((unused)))
{
   UInt16 error;

#ifdef DEBUG
   set_debug_traps();
   BREAKPOINT();
#endif

   switch(cmd)
   {
      case sysAppLaunchCmdNormalLaunch:
         error = CheckOsVersion(true);
         if(error)
         {
            if(GetROMVer(NULL, NULL) == 1)
               AppLaunchWithCommand(sysFileCDefaultApp,
                                    sysAppLaunchCmdNormalLaunch, NULL);
            return error;
         }

         error = StartApplication();
         if(error)
            return error;

         EventLoop();

         StopApplication();
         break;
   }

  return 0;
}

///////////////////////////
// Common Palm Functions //
///////////////////////////

static UInt16 CheckOsVersion(Boolean bDisplayAlert)
{
   UInt16 error = 0;

   // check if PalmOS 2.0 or greater
   if(GetROMVer(NULL, NULL) < 2)
   {
      if(bDisplayAlert)
         FrmCustomAlert(AlertOsVersionError, "2.0", NULL, NULL);
      error = sysErrRomIncompatible;
   }

   return error;
}

static UInt16 StartApplication(void)
{
   UInt32 version;
   UInt16 ROMVerMajor, ROMVerMinor;
   Boolean bPS, bSony;

   if(_TRGVGAFeaturePresent(&version))
   {
      bVgaExists = true;
      oldWinRect.extent.x = 240;
      oldWinRect.extent.y = 240;
   }

   // handle other DIA stuff
   bCollapsibleGraffiti = GetCollapseGraffiti(&bPS, &bSony);

   if(bCollapsibleGraffiti)
   {
      UInt16 cardNo;
      LocalID dbID;

      SysCurAppDatabase(&cardNo, &dbID);
      if(bPS)
      {
         SysNotifyRegister(cardNo, dbID, sysNotifyDisplayResizedEvent,
                           &MySysNotifyDisplayChangeEvent,
                           sysNotifyNormalPriority, NULL);
      }
      else if(bSony)
      {
         SysNotifyRegister(cardNo, dbID, sysNotifyDisplayChangeEvent,
                           &MySysNotifyDisplayChangeEvent,
                           sysNotifyNormalPriority, NULL);
      }
   }

   // load data
   LoadData();

   GetROMVer(&ROMVerMajor, &ROMVerMinor);
   realPowerOffAvail =
      (Boolean)(ROMVerMajor >= 4 || (ROMVerMajor == 3 && ROMVerMinor >= 2));

   if(bVgaExists)
   {
      VgaScreenModeType screenMode;

      VgaGetScreenMode(&screenMode, &currentRotateMode);
      VgaSetScreenMode(screenMode1To1, currentRotateMode);
   }

   FrmGotoForm(FormMain);

   return 0;
}

static void LoadData(void)
{
   Int16 prefVer;
   UInt16 prefSize;

   // get size of prefs
   prefSize = sizeof(data);
   prefVer = PrefGetAppPreferences(APP_ID, PREF_ID, &data, &prefSize, true);

   if(prefVer != PREF_VER)
   {
      data.asciiArmorKey = true;
      data.maskKey       = false;
      if(bVgaExists)
         data.font = VgaBaseToVgaFont(stdFont);
      else
         data.font = stdFont;
   }
}

static void EventLoop(void)
{
   UInt16 err;
   EventType e;
   UInt16 myPowerOffChr = (realPowerOffAvail ? vchrPowerOff : hardPowerChr);

   do
   {
      // wait for event or tiemout every second
      EvtGetEvent(&e, 100);

      // zero out data if power goes off
      if(e.eType == keyDownEvent)
         if(e.data.keyDown.modifiers & commandKeyMask)
            if(e.data.keyDown.chr == myPowerOffChr ||
               e.data.keyDown.chr == autoOffChr)
                  DoClearCmd(false);

      // get DIA updates sent to hidden form if necessary
      if((e.eType == winEnterEvent) && FrmGetActiveForm())
      {
         EventType newEvent;
	
         MemSet(&newEvent, sizeof(EventType), 0);
         newEvent.eType = winDisplayChangedEvent;
         EvtAddUniqueEventToQueue(&newEvent, 0, true);
      }

      // handle any system events
      if(SysHandleEvent(&e))
         continue;

      // handle any menu events (before I get messages)
      if(MenuHandleEvent(0, &e, &err))
         continue;

      // handle application events
      if(AppHandleEvent(&e))
         continue;

      // dispatch event to appropriate form
      FrmDispatchEvent(&e);

   }
   while(e.eType != appStopEvent);

   FrmCloseAllForms();
}

static Boolean AppHandleEvent(EventPtr event)
{
   Int32 formID;
   FormPtr pForm;

   // load appropriate form
   if(event->eType == frmLoadEvent)
   {
      formID = event->data.frmLoad.formID;
      pForm = FrmInitForm(formID);
      FrmSetActiveForm(pForm);
      switch(formID)
      {
         case FormMain:
            FrmSetEventHandler(pForm, MainEventHandler);
            break;
         default:
            ErrNonFatalDisplayIf(true, "Invalid Form Load Event");
            return true;
            break;
      }
   }

   return false;
}

static void StopApplication(void)
{
   Boolean bPS, bSony;

   // save data
   PrefSetAppPreferences(APP_ID, PREF_ID, PREF_VER, &data, sizeof(data), true);

   // clean up DIA stuff
   if(GetCollapseGraffiti(&bPS, &bSony))
   {
      UInt16 cardNo;
      LocalID dbID;

      SysCurAppDatabase(&cardNo, &dbID);
      if(bPS)
      {
         SysNotifyUnregister(cardNo, dbID, sysNotifyDisplayResizedEvent, NULL);
      }
      else if(bSony)
      {
         SysNotifyUnregister(cardNo, dbID, sysNotifyDisplayChangeEvent, NULL);
      }
   }
}

static struct cbuf *CbufNew(void)
{
   return (struct cbuf *)MemPtrNew(sizeof(struct cbuf));
}

static void CbufFree(struct cbuf *c)
{
   MemSet((void *)c, sizeof(struct cbuf), 0);
   MemPtrFree((void *)c);
}

static void ClearField(FieldPtr field)
{
   Char *p = FldGetTextPtr(field);
   UInt16 len;
   Char sTemp[80];

   if(p != 0)
   {
      len = FldGetTextLength(field);
      StrPrintF(sTemp, "len = %d", len);
      FldDelete(field, 0, len);
      MemSet((void *)p, len, 0);
   }
}

static struct cbuf *GetClipboard(void)
{
   struct cbuf *v;
   MemHandle clip;
   UInt16 len;
   UInt16 i;
   char *p;

   // returns a newly-allocated cbuf containing the clipboard
   // or 0 if the clipboard is empty

   if((clip = ClipboardGetItem(clipboardText, &len)) == 0 || len == 0)
   {
      return 0;
   }

   p = (unsigned char *)MemHandleLock(clip);
  
   v = CbufNew();

   for(i = 0; i < len; i++)
   {
      v->data[i] = p[i];
   }

   v->length = len;

   return v;
}

static UInt16 GetCurrentFieldId(void)
{
   FormPtr pForm = FrmGetActiveForm();
   UInt16 wFocused = FrmGetFocus(pForm);

   return FrmGetObjectId(pForm, wFocused);
}

static FieldPtr GetFieldOfActive(UInt16 field)
{
   FormPtr form;

   form = FrmGetActiveForm();
   return FrmGetObjectPtr(form, FrmGetObjectIndex(form, field));
}

static FieldPtr GetFocusObjectPtr(void)
{
   FormPtr frm;
   UInt16 focus;
	
   // Get a pointer to tha active form and the index of the form object
   // with focus.
   frm = FrmGetActiveForm();
   focus = FrmGetFocus(frm);
	
   // no object has the focus return NULL pointer.
   if(focus == noFocus)
      return(NULL);
		
   // Return a pointer to the object with focus.
   return(FrmGetObjectPtr(frm, focus));
}

static UInt16 GetROMVer(UInt16 *ROMVerMajor, UInt16 *ROMVerMinor)
{
   UInt32 dwROMVer;

   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &dwROMVer);

   if(ROMVerMajor != NULL)
      *ROMVerMajor = sysGetROMVerMajor(dwROMVer);
   if(ROMVerMinor != NULL)
      *ROMVerMinor = sysGetROMVerMinor(dwROMVer);

   return sysGetROMVerMajor(dwROMVer);
}

static void *Id2Ptr(UInt16 wID)
{
   return(FrmGetObjectPtr(FrmGetActiveForm(),
                          FrmGetObjectIndex(FrmGetActiveForm(), wID)));
}

static Err SendMe(Char *prefix)
{
   UInt16 cardNo;
   LocalID dbID;
   Err err;
   Char dbName[dmDBNameLength];
   Char *dbNamePRC = NULL;

   dbNamePRC = MemPtrNew(sizeof(Char) * (StrLen(prefix) + dmDBNameLength + 4));

   err = SysCurAppDatabase(&cardNo, &dbID);
   err = DmDatabaseInfo(cardNo, dbID, dbName, NULL, NULL, NULL, NULL, NULL,
                        NULL, NULL, NULL, NULL, NULL);
   StrPrintF(dbNamePRC, "%s%s.prc", prefix, dbName);
   err = SendDatabase(cardNo, dbID, dbNamePRC, dbName);

   // clean up
   if(dbNamePRC)
      MemPtrFree(dbNamePRC);

   return err;
}

static Err SendDatabase(UInt16 cardNo, LocalID dbID, Char *nameP,
                        Char *descriptionP)
{
   ExgSocketType exgSocket;
   Err err;

   // Create exgSocket structure
   MemSet(&exgSocket, sizeof(exgSocket), 0);
   exgSocket.description = descriptionP;
   exgSocket.name = nameP;

   // Start and exchange put operation
   err = ExgPut(&exgSocket);
   if(!err)
   {
      // This function converts a palm database into its external (public)
      // format. The first parameter is a callback that will be passed parts of
      // the database to send or write.
      err = ExgDBWrite(WriteDBData, &exgSocket, NULL, dbID, cardNo);
      // Disconnect Exg and pass error
      err = ExgDisconnect(&exgSocket, err);
   }

   return err;
}

static Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP)
{
   Err err;

   // Try to send as many bytes as were requested by the caller
   *sizeP = ExgSend((ExgSocketPtr)userDataP, (void*)dataP, *sizeP, &err);

   return err;
}

/////////////////////////
// Main Form Functions //
/////////////////////////

static Boolean MainEventHandler(EventPtr e)
{
   FormPtr pForm = FrmGetActiveForm();
   Boolean handled = true;

   // handle form specific messages
   switch(e->eType)
   {
      case frmOpenEvent:
         if(bVgaExists)
            VgaFormModify(pForm, vgaFormModify160To240);
         EditViewUpdateScrollBar();
         CtlSetValue(Id2Ptr(AsciiArmorKey), data.asciiArmorKey);
         CtlSetValue(Id2Ptr(MaskKey), data.maskKey);
         if(bCollapsibleGraffiti)
            CollapseSetUserState(pForm, true);
         if(bVgaExists)
            FldSetFont(Id2Ptr(FieldKey),
                       data.maskKey ? VgaBaseToVgaFont(symbol11Font) :
                                      VgaBaseToVgaFont(boldFont));
         else
            FldSetFont(Id2Ptr(FieldKey),
                       data.maskKey ? symbol11Font : boldFont);
         //FldSetFont(Id2Ptr(FieldKey), data.maskKey ? symbol11Font : boldFont);
         FldSetFont(Id2Ptr(FieldData), data.font);
         SizeForm(pForm);
         FrmDrawForm(pForm);
         FrmSetFocus(pForm, FrmGetObjectIndex(pForm, FieldKey));
         break;

      case displayExtentChangedEvent:
      case winDisplayChangedEvent:
         SizeForm(pForm);
         break;

      case frmCloseEvent:
         DoClearCmd(false);
         FrmEraseForm(pForm);
         FrmDeleteForm(pForm);
         break;

      case ctlSelectEvent:
         handled = ButtonCommand(e->data.ctlEnter.controlID);
         break;

      case menuEvent:
         handled = MenuCommand(e->data.menu.itemID);
         break;

      case menuOpenEvent:
         ChangeToGraffiti2MenuHelp();
         handled = false;
         break;

      case menuCmdBarOpenEvent:
         MenuCmdBarAddButton(menuCmdBarOnRight, BarBeamBitmap,
                             menuCmdBarResultMenuItem, CmdBeam, 0);
         MenuCmdBarAddButton(menuCmdBarOnRight, BarInfoBitmap,
                             menuCmdBarResultMenuItem, CmdInfo, 0);

         handled = false; // don't set handled to true; this event must fall
                          //   through to the system.
         break;

      case sclRepeatEvent:
         EditViewScroll(e->data.sclRepeat.newValue - 
                        e->data.sclRepeat.value, false);
         handled = false;
         break;

      case fldChangedEvent:
         EditViewUpdateScrollBar();
         handled = false;
         break;

      case keyDownEvent:
         switch(e->data.keyDown.chr)
         {
            case pageUpChr:
               Scroll(false);
               break;
            case pageDownChr:
               Scroll(true);
               break;
            case prevFieldChr:
               HandleNextPrevField(e->data.keyDown.chr);
               break;
            case nextFieldChr:
               HandleNextPrevField(e->data.keyDown.chr);
               break;
            case vchrSendData:
               SendMe(exgBeamPrefix);
               break;
            default:
               handled = false;
               break;
         }
         break;

      default:
         handled = false;
         break;
   }

   return handled;
}

static void HandleNextPrevField(UInt16 keyDown __attribute__ ((unused)))
{
   FormPtr pForm = FrmGetActiveForm();
   UInt16 fieldID = GetCurrentFieldId();

   switch(fieldID)
   {
      case FieldKey:
         fieldID = FieldData;
         break;

      case FieldData:
         fieldID = FieldKey;
         break;
   }
      
   FrmSetFocus(pForm, FrmGetObjectIndex(pForm, fieldID));
}

static void SetKey(key k)
{
   FieldPtr field;

   field = GetFieldOfActive(FieldKey);
   setkey(k, FldGetTextPtr(field), FldGetTextLength(field));
}

static void DoEncryptCmd(void)
{
   key k;
   struct cbuf *v;
   UInt16 len;
   UInt16 i;
   Char *p;
   FieldPtr field;

   field = GetFieldOfActive(FieldData);

   len = FldGetTextLength(field);
   if(len > MAX_PLAINTEXT_LEN)
   {
      FrmAlert(AlertTooBig);
      return;
   }

   p = FldGetTextPtr(field);
   if(p == 0)
   {
      return;
   }

   v = CbufNew();

   ((UInt32 *) v->data)[0] = TimGetSeconds();
   v->length = HEADER;

   for(i = 0; i < len; i++)
   {
      v->data[v->length++] = p[i];
   }

   SetKey(k);

   if(code(k, v) < 0)
   {
      FrmAlert(AlertTooBig);
   }
   else
   {
      ClipboardAddItem(clipboardText, v->data, v->length);
      FrmAlert(AlertEncryptOK);
   }

   CbufFree(v);

   if(data.asciiArmorKey)
      DoArmorCmd(true);

   return;
}

static Int16 SanityCheck(struct cbuf *v)
{
   Int16 i;

   // looks for obviously bogus data
   // at the moment this means anything containing nulls
   // somebody we may put on a fixed header

   for(i = HEADER; i < v->length; i++)
   {
      if(v->data[i] == 0)
         return -1;
   }

   return 0;
}

static void DoDecryptCmd(void)
{
   key k;
   struct cbuf *v;
   FieldPtr field;
   FormPtr pForm;

   if((v = GetClipboard()) == 0)
   {
      FrmAlert(AlertDecryptFailed);
      return;
   }

   SetKey(k);

   if(uncode(k, v) < 0 || SanityCheck(v) < 0)
   {
      FrmAlert(AlertDecryptFailed);
      CbufFree(v);
      return;
   }

   // else

   field = GetFieldOfActive(FieldData);
   ClearField(field);
   FldInsert(field, v->data + HEADER, v->length - HEADER);
   FldRecalculateField(field, true);
   CbufFree(v);

   // set the focus to the data field
   // for some reason the FieldKey has to have its focus set first
   pForm = FrmGetActiveForm();
   FrmSetFocus(pForm, FrmGetObjectIndex(pForm, FieldKey));
   FrmSetFocus(pForm, FrmGetObjectIndex(pForm, FieldData));

   // move to top of data field
   FldSetInsPtPosition(field, 0);

   return;
}

static void DoArmorCmd(Boolean calledFromEncrypt)
{
   struct cbuf *in, *out;

   if((in = GetClipboard()) == 0)
   {
      FrmAlert(AlertArmorNothing);
      return;
   }

   out = CbufNew();

   if(armor(in, out) == ERR_OVERFLOW)
   {
      FrmAlert(AlertArmorOverflow);
   }
   else
   {
      ClipboardAddItem(clipboardText, out->data, out->length);
      if(!calledFromEncrypt)
         FrmAlert(AlertArmorOK);
   }

   CbufFree(in);
   CbufFree(out);

   return;
}

static void DoClearCmd(Boolean askAboutClipboard)
{
   FormPtr pForm = FrmGetActiveForm();

   ClearField(GetFieldOfActive(FieldKey));
   ClearField(GetFieldOfActive(FieldData));

   FrmSetFocus(pForm, FrmGetObjectIndex(pForm, FieldKey));

   // ask if to clear the clipboard as well
   if(askAboutClipboard)
      if(FrmAlert(AlertClearClipboard) == 0)
         ClipboardAddItem(clipboardText, NULL, 0);
} 

static Boolean ButtonCommand(UInt16 command)
{
   Boolean handled = true;

   switch(command)
   {
      case CmdEncrypt:
         DoEncryptCmd();
         break;
      case CmdDecrypt:
         DoDecryptCmd();
         break;
      case CmdClear:
         DoClearCmd(true);
         break;
      case AsciiArmorKey:
         data.asciiArmorKey = CtlGetValue(Id2Ptr(AsciiArmorKey));
         break;
      case MaskKey:
         data.maskKey = CtlGetValue(Id2Ptr(MaskKey));
         if(bVgaExists)
            FldSetFont(Id2Ptr(FieldKey),
                       data.maskKey ? VgaBaseToVgaFont(symbol11Font) :
                                      VgaBaseToVgaFont(boldFont));
         else
            FldSetFont(Id2Ptr(FieldKey),
                       data.maskKey ? symbol11Font : boldFont);
         break;
      default:
         handled = false;
         break;
   }

   return handled;
}

static Boolean MenuCommand(UInt16 command)
{
   FieldPtr fld = GetFocusObjectPtr();
   Boolean handled = true;

   switch(command)
   {
      case CmdDecrypt:
         DoDecryptCmd();
         break;
      case CmdEncrypt:
         DoEncryptCmd();
         break;
      case CmdClear:
         DoClearCmd(true);
         break;
      case CmdUndo:
         if(fld) FldUndo(fld);
         break;
      case CmdCut:
         if(fld) FldCut(fld);
         break;
      case CmdCopy:
         if(fld) FldCopy(fld);
         break;
      case CmdPaste:
         if(fld) FldPaste(fld);
         break;
      case CmdSelectAll:
         if(fld) FldSetSelection(fld, 0, FldGetTextLength(fld));
         break;
      case CmdKeyboard:
         SysKeyboardDialog(kbdDefault);
         break;
      case CmdGraffitiHelp:
      case CmdGraffiti2Help:
         SysGraffitiReferenceDialog(referenceDefault);
         break;
      case CmdFont:
         data.font = FontSelect(data.font);
         FldSetFont(Id2Ptr(FieldData), data.font);
         EditViewUpdateScrollBar();
         break;
      case CmdBeam:
         SendMe(exgBeamPrefix);
         break;
      case CmdSend:
         SendMe(exgSendBeamPrefix);
         break;
      case CmdInfo:
         FrmHelp(StringInfo);
         break;
      case CmdAbout:
         DisplayAboutForm(FormAbout);
         break;
      case sysEditMenuUndoCmd:
      case sysEditMenuCutCmd:
      case sysEditMenuCopyCmd:
      case sysEditMenuPasteCmd:
         handled = false;
         break;
      default:
         ErrDisplayFileLineMsg("MenuCommand", command, "not handled");
         handled = false;
         break;
   }

   return handled;
}

static void Scroll(Boolean scrolldown)
{
   FieldPtr field;

   if((field = GetFieldOfActive(FieldData)) != 0)
   {
      UInt16 dataLines = FldGetVisibleLines(field) - 1;
      if(scrolldown)
      {
         if(FldScrollable(field, winDown))
         {
            FldScrollField(field, dataLines, winDown);
         }
      }
      else
      {
         if(FldScrollable(field, winUp))
         {
            FldScrollField(field, dataLines, winUp);
         }
      }
   }

   EditViewUpdateScrollBar();
}

static void ChangeToGraffiti2MenuHelp(void)
{
   UInt32 temp;
  
   // only Palm OS 5 or greater will have Graffiti 2
   if(GetROMVer(NULL, NULL) < 5)
      return;

   if(!FtrGet('grft', 1110, &temp) || !FtrGet('grf2', 1110, &temp))
   {
      Char sGraffiti2MenuText[40];

      MenuHideItem(CmdGraffitiHelp);
      SysCopyStringResource(sGraffiti2MenuText, StringGraffiti2Help);
      MenuAddItem(CmdKeyboard, CmdGraffiti2Help, 'G', sGraffiti2MenuText);
   }
}

//////////////////////////
// About Form Functions //
//////////////////////////

static void DisplayAboutForm(UInt16 rscID)
{
   FormPtr pForm = FrmGetActiveForm();
   FormPtr pNewForm = FrmInitForm(rscID);
   Boolean bFixScreen = false;

   if(bVgaExists)
      VgaFormModify(pNewForm, vgaFormModify160To240);

   // show the about form
   FrmDoDialog(pNewForm);

   if(bVgaExists && bFixScreen)
      VgaSetScreenMode(screenMode1To1, currentRotateMode);

   FrmDeleteForm(pNewForm);
   FrmSetActiveForm(pForm);
}

/////////////////////////
// Scrollbar Functions //
/////////////////////////

static void EditViewUpdateScrollBar(void)
{
   UInt16 scrollPos;
   UInt16 textHeight;
   UInt16 fieldHeight;
   Int16 maxValue;
   FieldPtr pField;
   ScrollBarPtr bar;

   pField = Id2Ptr(FieldData);
   bar = Id2Ptr(ScrollBar);
	
   FldGetScrollValues(pField, &scrollPos, &textHeight,  &fieldHeight);

   if(textHeight > fieldHeight)
   {
      // On occasion, such as after deleting a multi-line selection of text,
      // the display might be the last few lines of a field followed by some
      // blank lines.  To keep the current position in place and allow the user
      // to "gracefully" scroll out of the blank area, the number of blank lines
      // visible needs to be added to max value.  Otherwise the scroll position
      // may be greater than maxValue, get pinned to maxvalue in SclSetScrollBar
      // resulting in the scroll bar and the display being out of sync.
      maxValue = (textHeight - fieldHeight) + FldGetNumberOfBlankLines(pField);
   }
   else if(scrollPos)
      maxValue = scrollPos;
   else
      maxValue = 0;
		
   SclSetScrollBar(bar, scrollPos, 0, maxValue, fieldHeight-1);
}

static void EditViewScroll(Int16 linesToScroll, Boolean updateScrollbar)
{
   UInt16 blankLines;
   FieldPtr pField;
	
   pField = Id2Ptr(FieldData);
   blankLines = FldGetNumberOfBlankLines(pField);

   if(linesToScroll < 0)
      FldScrollField(pField, -linesToScroll, winUp);
   else if(linesToScroll > 0)
      FldScrollField(pField, linesToScroll, winDown);

   // If there were blank lines visible at the end of the field
   // then we need to update the scroll bar.
   if(blankLines || updateScrollbar)
   {
      ErrNonFatalDisplayIf(blankLines && linesToScroll > 0,
                           "blank lines when scrolling winDown");
		
      EditViewUpdateScrollBar();
   }
}

////////////////////////////////////
// HandEra & DIA Resize Functions //
////////////////////////////////////

static void SizeForm(FormPtr pForm)
{
   RectangleType winRect;
   Coord xDiff, yDiff;
   Boolean bDraw;

   if(!bVgaExists && !bCollapsibleGraffiti)
      return;

   bDraw = FrmVisible(pForm);

   // set new window bounds
   winRect.topLeft.x = 0;
   winRect.topLeft.y = 0;
   WinGetDisplayExtent(&winRect.extent.x, &winRect.extent.y);
   xDiff = winRect.extent.x - oldWinRect.extent.x;
   yDiff = winRect.extent.y - oldWinRect.extent.y;
   oldWinRect = winRect;

   WinSetWindowBounds(FrmGetWindowHandle(pForm), &winRect);

   if((xDiff == 0) && (yDiff == 0))
      return;

   // move and size objects on form
   ResizeObject(pForm, FrmGetObjectIndex(pForm, MaskKey),
                xDiff, 0, xDiff, 0, false, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, AsciiArmorKey),
                xDiff, 0, xDiff, 0, false, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, FieldKey),
                0, 0, xDiff, 0, bDraw && bVgaExists, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, FieldData),
                0, 0, xDiff, yDiff, bDraw && bVgaExists, true);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, CmdDecrypt),
                0, yDiff, 0, 0, false, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, CmdEncrypt),
                0, yDiff, 0, 0, false, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, CmdClear),
                0, yDiff, 0, 0, false, false);
   ResizeObject(pForm, FrmGetObjectIndex(pForm, ScrollBar),
                xDiff, 0, 0, yDiff, bDraw && bVgaExists, false);
   EditViewUpdateScrollBar();
   ResizeObject(pForm, GetGSIIndex(pForm),
                0, yDiff, 0, 0, bDraw && bVgaExists, false);

   // redraw form (make sure graffiti shift indicator gets drawn)
   if(bDraw)
      FrmDrawForm(pForm);
}

static UInt16 GetGSIIndex(FormPtr pForm)
{
   UInt16 i, numObjects;

   numObjects = FrmGetNumberOfObjects(pForm);
   for(i = 0; i < numObjects; i++)
   {
      if(FrmGetObjectType(pForm, i) == frmGraffitiStateObj)
         return i;
   }

   return -1;
}

static void ResizeObject(FormPtr pForm, UInt16 objectIndex,
                         Coord xMove, Coord yMove,
                         Coord xExp, Coord yExp, Boolean bDraw,
                         Boolean bMultiLineField)
{
   RectangleType r;
   FieldPtr pField = NULL;
   UInt16 tmpPos = 0xFFFF, tmpStart, tmpEnd;
    
   if((xMove == 0) && (yMove == 0) && (xExp == 0) && (yExp == 0))
      return;
    
   if(bDraw && bMultiLineField)
   {
      pField = FrmGetObjectPtr(pForm, objectIndex);
      FldGetSelection(pField, &tmpStart, &tmpEnd);
      if(InsPtEnabled() || (tmpStart != tmpEnd))
         tmpPos = FldGetInsPtPosition(pField);
   }

   FrmGetObjectBounds(pForm, objectIndex, &r);
   if(bDraw)
      FrmHideObject(pForm, objectIndex);
   r.topLeft.x += xMove;
   r.topLeft.y += yMove;
   r.extent.x  += xExp;
   r.extent.y  += yExp;
   FrmSetObjectBounds(pForm, objectIndex, &r);
    
   if(FrmGetObjectType(pForm, objectIndex) == frmGraffitiStateObj)
      GsiSetLocation(r.topLeft.x, r.topLeft.y);
    
   if(bDraw && bMultiLineField)
   {
      if(tmpPos != 0xFFFF)
      {
         FldSetInsPtPosition(pField, tmpPos);
         FldSendChangeNotification(pField);
         FldSetSelection(pField, tmpStart, tmpEnd);
      }
   }

   if(bDraw)
      FrmShowObject(pForm, objectIndex);
}

static void CollapseSetUserState(FormPtr pForm, Boolean bForce)
{
   WinHandle hWin = WinGetWindowHandle(pForm);
   Boolean bPS;

   GetCollapseGraffiti(&bPS, NULL);

   if(bPS)
   {
      RectangleType frmBounds;
      Coord x, y; 

      FrmGetFormBounds(pForm, &frmBounds);
      y = frmBounds.extent.y;
      x = frmBounds.extent.x;

      if(bForce || (PINGetInputTriggerState() == pinInputTriggerEnabled))
      {
         WinSetConstraintsSize(hWin, y, pinMaxConstraintSize,
                               pinMaxConstraintSize, x, x,
                               pinMaxConstraintSize);
         FrmSetDIAPolicyAttr(pForm, frmDIAPolicyCustom);
         PINSetInputTriggerState(pinInputTriggerEnabled);
         PINSetInputAreaState(pinInputAreaUser); 
      }
   }
}

static Boolean GetCollapseGraffiti(Boolean *pbPS, Boolean *pbSony)
{
   UInt32 pinMgrVersion;
   Boolean bPS   = false;
   Boolean bSony = false;
   Err error;
   UInt16 refNum;

   // PalmSource detection
   error = FtrGet(pinCreator, pinFtrAPIVersion, &pinMgrVersion);
   if(error == errNone)
   {
      if(pinMgrVersion == pinAPIVersion1_0)
      {
         // make sure that it works with the Gramin iQue 3600
         bPS = true;
      }
      else if(pinMgrVersion >= pinAPIVersion1_1)
      {
         UInt32 flags;

         error = FtrGet(sysFtrCreator, sysFtrNumInputAreaFlags, &flags);
         if((error == errNone) && (flags & grfFtrInputAreaFlagCollapsible))
         {
            bPS = true;
         }
      }
   }

   // Sony detection
   if(SysLibFind(sonySysLibNameSilk, &refNum) != sysErrLibNotFound)
   {
      if(SilkLibOpen(refNum) == errNone)
      {
         SilkLibEnableResize(refNum);
         bSony = true;
      }
   }

   // return values
   if(pbPS)
      *pbPS = bPS;
   if(pbSony)
      *pbSony = bSony;

   return (bPS || bSony);
}

static Err MySysNotifyDisplayChangeEvent(SysNotifyParamType *np
                                         __attribute__ ((unused)))
{
   EventType newEvent;
	
   MemSet(&newEvent, sizeof(EventType), 0);

   newEvent.eType = winDisplayChangedEvent;
	
   EvtAddUniqueEventToQueue(&newEvent, 0, true);
	
   return 0;
}
